"""Support for loading OpenDRIVE maps.

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree:

   workspace
   xodr_parser
"""

from .workspace import OpenDriveWorkspace
from .xodr_parser import OpenDriveWarning




"""Support for lazy evaluation of expressions and specifiers."""

import itertools
import types

from scenic.core.utils import DefaultIdentityDict

class LazilyEvaluable:
	"""Values which may require evaluation in the context of an object being constructed.

	If a LazilyEvaluable specifies any properties it depends on, then it cannot be evaluated to a
	normal value except during the construction of an object which already has values for those
	properties.
	"""
	def __init__(self, requiredProps):
		self._dependencies = ()		# TODO improve?
		self._requiredProperties = set(requiredProps)

	def evaluateIn(self, context):
		"""Evaluate this value in the context of an object being constructed.

		The object must define all of the properties on which this value depends.
		"""
		cache = context._evaluated		# cache of lazy values already evaluated in this context
		if self in cache:
			return cache[self]	# avoid making a new evaluated copy of this value
		assert all(hasattr(context, prop) for prop in self._requiredProperties)
		value = self.evaluateInner(context)
		assert not needsLazyEvaluation(value)	# value should not require further evaluation
		cache[self] = value
		return value

	def evaluateInner(self, context):
		"""Actually evaluate in the given context, which provides all required properties."""
		return self

	@staticmethod
	def makeContext(**props):
		"""Make a context with the given properties for testing purposes."""
		context = types.SimpleNamespace(**props)
		context._evaluated = DefaultIdentityDict()
		return context

class DelayedArgument(LazilyEvaluable):
	"""Specifier arguments requiring other properties to be evaluated first.

	The value of a DelayedArgument is given by a function mapping the context (object under
	construction) to a value.
	"""
	def __new__(cls, *args, _internal=False, **kwargs):
		darg = super().__new__(cls)
		if _internal:
			return darg
		# at runtime, evaluate immediately in the context of the current agent
		import scenic.syntax.veneer as veneer
		if veneer.simulationInProgress() and veneer.currentBehavior:
			agent = veneer.currentBehavior.agent
			assert agent
			darg.__init__(*args, **kwargs)
			agent._evaluated = DefaultIdentityDict()
			value = darg.evaluateIn(agent)
			del agent._evaluated
			return value
		else:
			return darg

	def __init__(self, requiredProps, value, _internal=False):
		self.value = value
		super().__init__(requiredProps)

	def evaluateInner(self, context):
		return self.value(context)

	def __getattr__(self, name):
		return DelayedArgument(self._requiredProperties,
			lambda context: getattr(self.evaluateIn(context), name))

	def __call__(self, *args, **kwargs):
		subprops = (requiredProperties(arg) for arg in itertools.chain(args, kwargs.values()))
		props = self._requiredProperties.union(*subprops)
		def value(context):
			subvalues = (valueInContext(arg, context) for arg in args)
			kwsvs = { name: valueInContext(arg, context) for name, arg in kwargs.items() }
			return self.evaluateIn(context)(*subvalues, **kwsvs)
		return DelayedArgument(props, value)

# Operators which can be applied to DelayedArguments
allowedOperators = [
	'__neg__',
	'__pos__',
	'__abs__',
	'__lt__', '__le__',
	'__eq__', '__ne__',
	'__gt__', '__ge__',
	'__add__', '__radd__',
	'__sub__', '__rsub__',
	'__mul__', '__rmul__',
	'__truediv__', '__rtruediv__',
	'__floordiv__', '__rfloordiv__',
	'__mod__', '__rmod__',
	'__divmod__', '__rdivmod__',
	'__pow__', '__rpow__',
	'__round__',
	'__len__',
	'__getitem__'
]
def makeDelayedOperatorHandler(op):
	def handler(self, *args):
		props = self._requiredProperties.union(*(requiredProperties(arg) for arg in args))
		def value(context):
			subvalues = (valueInContext(arg, context) for arg in args)
			return getattr(self.evaluateIn(context), op)(*subvalues)
		return DelayedArgument(props, value)
	return handler
for op in allowedOperators:
	setattr(DelayedArgument, op, makeDelayedOperatorHandler(op))

def makeDelayedFunctionCall(func, args, kwargs):
	"""Utility function for creating a lazily-evaluated function call."""
	props = set().union(*(requiredProperties(arg)
	                      for arg in itertools.chain(args, kwargs.values())))
	def value(context):
		subvalues = (valueInContext(arg, context) for arg in args)
		kwsubvals = { name: valueInContext(arg, context) for name, arg in kwargs.items() }
		return func(*subvalues, **kwsubvals)
	return DelayedArgument(props, value)

def valueInContext(value, context):
	"""Evaluate something in the context of an object being constructed."""
	if isinstance(value, LazilyEvaluable):
		return value.evaluateIn(context)
	else:
		return value

def requiredProperties(thing):
	if isinstance(thing, LazilyEvaluable):
		return thing._requiredProperties
	else:
		return set()

def needsLazyEvaluation(thing):
	return isinstance(thing, DelayedArgument) or requiredProperties(thing)



"""Various utilities for working with Webots scenarios."""

import math

import numpy as np

from scenic.core.geometry import normalizeAngle

def webotsToScenicPosition(pos):
    """Convert a Webots position to a Scenic position.

    Drops the Webots Y coordinate.
    """
    x, y, z = pos
    return (x, -z)

def scenicToWebotsPosition(pos, y=0):
    """Convert a Scenic position to a Webots position."""
    x, z = pos
    return [x, y, -z]

def webotsToScenicRotation(rot, tolerance2D=None):
    """Convert a Webots rotation vector to a Scenic heading.

    Assumes the object lies in the Webots X-Z plane, with a rotation axis
    close to the Y axis. If ``tolerance2D`` is given, returns ``None`` if the
    orientation of the object is not sufficiently close to being 2D.
    """
    axis = np.array(rot[:3])
    angle = rot[3]
    if tolerance2D is not None and np.linalg.norm(axis - (0, 1, 0)) > tolerance2D:
        return None
    return normalizeAngle(angle + math.pi)

def scenicToWebotsRotation(heading):
    """Convert a Scenic heading to a Webots rotation vector."""
    return [0, 1, 0, heading - math.pi]




### Top-level functionality of the scenic package as a script:
### load a scenario and generate scenes in an infinite loop.

import sys
import time
import argparse
import random

if sys.version_info >= (3, 8):
    from importlib import metadata
else:
    import importlib_metadata as metadata

import scenic.syntax.translator as translator
import scenic.core.errors as errors
from scenic.core.simulators import SimulationCreationError

parser = argparse.ArgumentParser(prog='scenic', add_help=False,
                                 usage='scenic [-h | --help] [options] FILE [options]',
                                 description='Sample from a Scenic scenario, optionally '
                                             'running dynamic simulations.')

mainOptions = parser.add_argument_group('main options')
mainOptions.add_argument('-S', '--simulate', action='store_true',
                         help='run dynamic simulations from scenes '
                              'instead of simply showing diagrams of scenes')
mainOptions.add_argument('-s', '--seed', help='random seed', type=int)
mainOptions.add_argument('-v', '--verbosity', help='verbosity level (default 1)',
                         type=int, choices=(0, 1, 2, 3), default=1)
mainOptions.add_argument('-p', '--param', help='override a global parameter',
                         nargs=2, default=[], action='append', metavar=('PARAM', 'VALUE'))
mainOptions.add_argument('-m', '--model', help='specify a Scenic world model', default=None)
mainOptions.add_argument('--scenario', default=None,
                         help='name of scenario to run (if file contains multiple)')

# Simulation options
simOpts = parser.add_argument_group('dynamic simulation options')
simOpts.add_argument('--time', help='time bound for simulations (default none)',
                     type=int, default=None)
simOpts.add_argument('--count', help='number of successful simulations to run (default infinity)',
                     type=int, default=0)
simOpts.add_argument('--max-sims-per-scene', type=int, default=1, metavar='N',
                     help='max # of rejected simulations before sampling a new scene (default 1)')

# Interactive rendering options
intOptions = parser.add_argument_group('static scene diagramming options')
intOptions.add_argument('-d', '--delay', type=float,
                        help='loop automatically with this delay (in seconds) '
                             'instead of waiting for the user to close the diagram')
intOptions.add_argument('-z', '--zoom', help='zoom expansion factor (default 1)',
                        type=float, default=1)

# Debugging options
debugOpts = parser.add_argument_group('debugging options')
debugOpts.add_argument('--show-params', help='show values of global parameters',
                       action='store_true')
debugOpts.add_argument('--show-records', help='show values of recorded expressions',
                       action='store_true')
debugOpts.add_argument('-b', '--full-backtrace', help='show full internal backtraces',
                       action='store_true')
debugOpts.add_argument('--pdb', action='store_true',
                       help='enter interactive debugger on errors (implies "-b")')
ver = metadata.version('scenic')
debugOpts.add_argument('--version', action='version', version=f'Scenic {ver}',
                       help='print Scenic version information and exit')
debugOpts.add_argument('--dump-initial-python', help='dump initial translated Python',
                       action='store_true')
debugOpts.add_argument('--dump-ast', help='dump final AST', action='store_true')
debugOpts.add_argument('--dump-python', help='dump Python equivalent of final AST',
                       action='store_true')
debugOpts.add_argument('--no-pruning', help='disable pruning', action='store_true')
debugOpts.add_argument('--gather-stats', type=int, metavar='N',
                       help='collect timing statistics over this many scenes')

parser.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS,
                    help=argparse.SUPPRESS)

# Positional arguments
parser.add_argument('scenicFile', help='a Scenic file to run', metavar='FILE')

# Parse arguments and set up configuration
args = parser.parse_args()
delay = args.delay
errors.showInternalBacktrace = args.full_backtrace
if args.pdb:
    errors.postMortemDebugging = True
    errors.showInternalBacktrace = True
params = {}
for name, value in args.param:
    # Convert params to ints or floats if possible
    try:
        value = int(value)
    except ValueError:
        try:
            value = float(value)
        except ValueError:
            pass
    params[name] = value
translator.dumpTranslatedPython = args.dump_initial_python
translator.dumpFinalAST = args.dump_ast
translator.dumpASTPython = args.dump_python
translator.verbosity = args.verbosity
translator.usePruning = not args.no_pruning
if args.seed is not None and args.verbosity >= 1:
    print(f'Using random seed = {args.seed}')
    random.seed(args.seed)

# Load scenario from file
if args.verbosity >= 1:
    print('Beginning scenario construction...')
startTime = time.time()
scenario = errors.callBeginningScenicTrace(
    lambda: translator.scenarioFromFile(args.scenicFile,
                                        params=params,
                                        model=args.model,
                                        scenario=args.scenario)
)
totalTime = time.time() - startTime
if args.verbosity >= 1:
    print(f'Scenario constructed in {totalTime:.2f} seconds.')

if args.simulate:
    simulator = errors.callBeginningScenicTrace(scenario.getSimulator)

def generateScene():
    startTime = time.time()
    scene, iterations = errors.callBeginningScenicTrace(
        lambda: scenario.generate(verbosity=args.verbosity)
    )
    if args.verbosity >= 1:
        totalTime = time.time() - startTime
        print(f'  Generated scene in {iterations} iterations, {totalTime:.4g} seconds.')
        if args.show_params:
            for param, value in scene.params.items():
                print(f'    Parameter "{param}": {value}')
    return scene, iterations

def runSimulation(scene):
    startTime = time.time()
    if args.verbosity >= 1:
        print(f'  Beginning simulation of {scene.dynamicScenario}...')
    try:
        simulation = errors.callBeginningScenicTrace(
            lambda: simulator.simulate(scene, maxSteps=args.time, verbosity=args.verbosity,
                                       maxIterations=args.max_sims_per_scene)
        )
    except SimulationCreationError as e:
        if args.verbosity >= 1:
            print(f'  Failed to create simulation: {e}')
        return False
    if args.verbosity >= 1:
        totalTime = time.time() - startTime
        print(f'  Ran simulation in {totalTime:.4g} seconds.')
    if simulation and args.show_records:
        for name, value in simulation.result.records.items():
            if isinstance(value, list):
                print(f'    Record "{name}": (time series)')
                for step, subval in value:
                    print(f'      {step:4d}: {subval}')
            else:
                print(f'    Record "{name}": {value}')
    return simulation is not None

try:
    if args.gather_stats is None:   # Generate scenes interactively until killed
        import matplotlib.pyplot as plt
        successCount = 0
        while True:
            scene, _ = generateScene()
            if args.simulate:
                success = runSimulation(scene)
                if success:
                    successCount += 1
                    if 0 < args.count <= successCount:
                        break
            else:
                if delay is None:
                    scene.show(zoom=args.zoom)
                else:
                    scene.show(zoom=args.zoom, block=False)
                    plt.pause(delay)
                    plt.clf()
    else:   # Gather statistics over the specified number of scenes
        its = []
        startTime = time.time()
        while len(its) < args.gather_stats:
            scene, iterations = generateScene()
            its.append(iterations)
        totalTime = time.time() - startTime
        count = len(its)
        print(f'Sampled {len(its)} scenes in {totalTime:.2f} seconds.')
        print(f'Average iterations/scene: {sum(its)/count}')
        print(f'Average time/scene: {totalTime/count:.2f} seconds.')

except KeyboardInterrupt:
    pass

finally:
    if args.simulate:
        simulator.destroy()

def dummy():    # for the 'scenic' entry point to call after importing this module
    pass



"""Actions for dynamic agents in Webots simulations."""

import pickle

from scenic.core.simulators import Action
import scenic.simulators.webots.utils as utils

class OffsetAction(Action):
    """Move an object by the given offset relative to its current heading."""
    def __init__(self, offset):
        self.offset = offset

    def applyTo(self, obj, webotsObj):
        pos = obj.position.offsetRotated(obj.heading, self.offset)
        pos = utils.scenicToWebotsPosition(pos, y=obj.elevation)
        webotsObj.getField('translation').setSFVec3f(pos)

class WriteFileAction(Action):
    """Pickle the given data and write the result to a file.

    For use in communication with external controllers or other code.
    """
    def __init__(self, path, data):
        self.path = path
        self.data = data

    def applyTo(self, obj, webotsObj):
        with open(self.path, 'wb') as outFile:
            pickle.dump(self.data, outFile)



"""Utilities to help serialize Scenic objects."""

from scenic.core.vectors import Vector

def scenicToJSON(obj):
    """Utility function to help serialize Scenic objects to JSON.

    Suitable for passing as the ``default`` argument to `json.dump`.
    """
    if isinstance(obj, Vector):
        return list(obj)
    raise TypeError(f'Object of type {obj.__class__.__name__} is not JSON serializable')



"""World models and interfaces for particular simulators.

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree:

   newtonian
   carla
   gta
   lgsvl
   webots
   xplane
   utils
"""



import math

from scenic.core.vectors import Vector
from scenic.core.geometry import normalizeAngle


def vectorFromHeading(heading):
	return Vector(0, 1).rotatedBy(heading)

def angularSpeedFromSteer(steer, speed, length):
	return speed * math.tan(steer) / length

def headingFromVector(vec):
	x, y = vec.x, vec.y
	if x == 0 and y == 0:
		return None
	return math.atan2(y, x) - math.pi / 2


"""Scenic world models for the Webots robotics simulator.

This module contains common code for working with Webots, e.g. parsing WBT files,
as well as a generic dynamic simulator interface and world model for Webots.
More detailed world models for particular types of scenarios are in submodules.

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree:

   simulator
   model
   actions
   mars
   road
   guideways
   utils
   world_parser
"""

from .utils import scenicToWebotsPosition, scenicToWebotsRotation
from .simulator import WebotsSimulator



"""Simple Newtonian physics simulator for traffic scenarios.

Allows scenarios written using the :obj:`scenic.domains.driving` abstract
domain to be simulated without installing an external simulator.

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree:

   model
   simulator
"""


"""Newtonian simulator implementation."""

import numpy as np
import math
from math import sin, radians, degrees, copysign
import shapely
import os

from scenic.domains.driving.simulators import DrivingSimulator, DrivingSimulation
from scenic.core.simulators import SimulationCreationError
from scenic.syntax.veneer import verbosePrint
from scenic.core.vectors import Vector
import scenic.simulators.newtonian.utils.utils as utils
from scenic.domains.driving.roads import Network
from scenic.syntax.translator import verbosity
if verbosity == 0:	# suppress pygame advertisement at zero verbosity
	import os
	os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = 'hide'
import pygame
import time

import pathlib
current_dir = pathlib.Path(__file__).parent.absolute()

WIDTH = 1280
HEIGHT = 800
MAX_ACCELERATION = 5.6 # in m/s2, seems to be a pretty reasonable value
MAX_BRAKING = 4.6

class NewtonianSimulator(DrivingSimulator):
    """Implementation of `Simulator` for the Newtonian simulator."""
    def __init__(self, network, timestep=0.1, render=False):
        self.timestep = timestep
        self.render = render
        self.network = network

    def createSimulation(self, scene, verbosity=0):
        return NewtonianSimulation(scene, self.network, timestep=self.timestep, verbosity=verbosity, render=self.render)

class NewtonianSimulation(DrivingSimulation):
    def __init__(self, scene, network, timestep, verbosity=0, render=False):
        super().__init__(scene, timestep=timestep, verbosity=verbosity)
        self.render = render
        self.network = network
        self.ego = self.objects[0]

        # Set actor's initial velocity (if specified)
        for obj in self.objects:
            if obj.speed is not None:
                equivVel = obj.speed*utils.vectorFromHeading(obj.heading)
                obj.setVelocity(equivVel)

        if self.render:
            # determine window size
            min_x, min_y = self.ego.position
            max_x, max_y = self.ego.position
            for obj in self.objects:
                x, y = obj.position
                min_x, max_x = min(min_x, x), max(max_x, x)
                min_y, max_y = min(min_y, y), max(max_y, y)

            pygame.init()
            pygame.font.init()
            self.screen = pygame.display.set_mode((WIDTH,HEIGHT), pygame.HWSURFACE | pygame.DOUBLEBUF)
            self.screen.fill((255, 255, 255))
            x, y = self.ego.position
            self.x_window = [min_x-50, max_x+50]
            self.y_window = [min_y-50, max_y+50]
            xlim1, xlim2 = self.x_window
            ylim1, ylim2 = self.y_window
            img_path = os.path.join(current_dir, 'car.png')
            self.car = pygame.image.load(img_path)
            self.car_width = int(3.5 * WIDTH / (xlim2 - xlim1))
            self.car_height = self.car_width
            self.car = pygame.transform.scale(self.car, (self.car_width, self.car_height))
            self.parse_network()
            self.draw_objects()

    def parse_network(self):
        self.network_polygons = []
        for element in self.network.elements.values():
            if not hasattr(element, 'polygon'):
                continue
            if isinstance(element.polygon, shapely.geometry.multipolygon.MultiPolygon):
                all_polygons = list(element.polygon)
            else:
                all_polygons = [element.polygon]
            for poly in all_polygons:
                if self.boundingBoxOnScreen(*poly.bounds):
                    screenPoints = map(self.scenicToScreenVal, poly.exterior.coords)
                    self.network_polygons.append(list(screenPoints))

    def scenicToScreenVal(self, pos):
        x, y = pos
        min_x, max_x = self.x_window
        min_y, max_y = self.y_window
        x_prop = (x - min_x) / (max_x - min_x)
        y_prop = (y - min_y) / (max_y - min_y)
        return int(x_prop * WIDTH), HEIGHT - 1 - int(y_prop * HEIGHT)

    def actionsAreCompatible(self, agent, actions):
        return True

    def boundingBoxOnScreen(self, x1, y1, x2, y2):
        min_x, max_x = self.x_window
        min_y, max_y = self.y_window
        onScreen = lambda x, y: min_x <= x <= max_x and \
                                min_y <= y <= max_y
        return onScreen(x1, y1) or \
               onScreen(x2, y2) or \
               onScreen(x1, y2) or \
               onScreen(x2, y1)

    def step(self):
        for obj in self.objects:
            if obj.hand_brake:
                acceleration = -MAX_BRAKING
            elif obj.brake > 0:
                acceleration = -obj.brake * MAX_BRAKING
            else:
                acceleration = obj.throttle * MAX_ACCELERATION
            obj.speed += acceleration * self.timestep
            obj.velocity = Vector(0, obj.speed).rotatedBy(obj.heading)
            if obj.steer:
                turning_radius = obj.length / sin(obj.steer * np.pi / 2)
                angular_velocity = obj.speed / turning_radius
            else:
                angular_velocity = 0
            obj.position += obj.velocity * self.timestep
            obj.heading -= angular_velocity * self.timestep
        if self.render:
            self.draw_objects()

    def draw_objects(self):
        self.screen.fill((255, 255, 255))
        for screenPoints in self.network_polygons:
            pygame.draw.polygon(self.screen, (0, 0, 0), list(screenPoints), width=1)

        for obj in self.objects:
            color = (255, 0, 0) if obj is self.ego else (0, 0, 255)
            h, w = obj.length, obj.width
            pos_vec = Vector(-1.75, 1.75)
            neg_vec = Vector(w / 2, h / 2)
            heading_vec = Vector(0, 10).rotatedBy(obj.heading)
            dx, dy = int(heading_vec.x), -int(heading_vec.y)
            x, y = self.scenicToScreenVal(obj.position)
            rect_x, rect_y = self.scenicToScreenVal(obj.position + pos_vec)
            self.rotated_car = pygame.transform.rotate(self.car, obj.heading * 180 / np.pi)
            self.screen.blit(self.rotated_car, (rect_x, rect_y))
        pygame.display.update()
        time.sleep(self.timestep)

    def getProperties(self, obj, properties):
        values = dict(
            position=obj.position,
            elevation=obj.elevation,
            heading=obj.heading,
            velocity=obj.velocity,
            speed=obj.speed,
            angularSpeed=obj.angularSpeed,
        )
        return values
